from aerosandbox import ExplicitAnalysis
from aerosandbox.geometry import *
from aerosandbox.performance import OperatingPoint
import aerosandbox.library.aerodynamics as aero
import aerosandbox.numpy as np
from aerosandbox.aerodynamics.aero_3D.aero_buildup_submodels.fuselage_aerodynamics_utilities import *
from aerosandbox.library.aerodynamics import transonic
import aerosandbox.library.aerodynamics as aerolib


class AeroBuildup(ExplicitAnalysis):
    """
    A workbook-style aerodynamics buildup.
    """
    default_analysis_specific_options = {
        Airplane: dict(),
        Wing    : dict(
            additional_CL=0,
            additional_CD=0,
            additional_CM=0,
            additional_L=0,
            additional_D=0,
            additional_M=0,
        ),
        WingXSec: dict(
            additional_CL=0,
            additional_CD=0,
            additional_CM=0,
            additional_L=0,
            additional_D=0,
            additional_M=0,
        ),
        Fuselage: dict(
            E_wave_drag=2.5,  # Wave drag efficiency factor

            # Defined by Raymer, "Aircraft Design: A Conceptual Approach", 2nd Ed. Chap. 12.5.9 "Supersonic Parasite Drag".
            # Notated there as "E_WD".
            #
            # Various recommendations:
            #   * For a perfect Sears-Haack body, 1.0
            #   * For a clean aircraft with smooth volume distribution (e.g., BWB), 1.2
            #   * For a "more typical supersonic...", 1.8 - 2.2
            #   * For a "poor supersonic design", 2.5 - 3.0
            #   * The F-15 has E_WD = 2.9.

            nose_fineness_ratio=3,  # Fineness ratio (length / diameter) of the nose section of the fuselage.

            # Impacts wave drag calculations, among other things.

            additional_CL=0,
            additional_CD=0,
            additional_CM=0,
            additional_L=0,
            additional_D=0,
            additional_M=0,
        ),
    }

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 include_wave_drag: bool = True,
                 ):
        ### Initialize
        self.airplane = airplane
        self.op_point = op_point
        self.include_wave_drag = include_wave_drag

    def run(self):
        ### Compute the forces on each component
        aero_components = [
                              self.wing_aerodynamics(wing=wing.translate(-self.airplane.xyz_ref)) for wing in
                              self.airplane.wings
                          ] + [
                              self.fuselage_aerodynamics(fuselage=fuse.translate(-self.airplane.xyz_ref)) for fuse in
                              self.airplane.fuselages
                          ]

        ### Sum up the forces
        aero_total = {}

        for k in aero_components[0].keys():
            values = [
                component[k] for component in aero_components
            ]

            try:
                aero_total[k] = sum(values)
            except TypeError:
                aero_total[k] = [
                    sum([
                        value[i]
                        for value in values
                    ])
                    for i in range(3)
                ]

        ##### Add nondimensional forces, and nondimensional quantities.
        if self.airplane.s_ref is not None:
            qS = self.op_point.dynamic_pressure() * self.airplane.s_ref

            aero_total["CL"] = aero_total["L"] / qS
            aero_total["CY"] = aero_total["Y"] / qS
            aero_total["CD"] = aero_total["D"] / qS
            aero_total["Cl"] = aero_total["l_b"] / qS / self.airplane.b_ref
            aero_total["Cm"] = aero_total["m_b"] / qS / self.airplane.c_ref
            aero_total["Cn"] = aero_total["n_b"] / qS / self.airplane.b_ref

            self.output = aero_total

        return aero_total

    def wing_aerodynamics(self,
                          wing: Wing,
                          ):
        """
        Estimates the aerodynamic forces, moments, and derivatives on a wing in isolation.

        Moments are given with the reference at Wing [0, 0, 0].

        Args:

            wing: A Wing object that you wish to analyze.

            op_point: The OperatingPoint that you wish to analyze the fuselage at.

        TODO account for wing airfoil pitching moment

        Returns:

        """
        ##### Alias a few things for convenience
        op_point = self.op_point
        wing_options = self.get_options(wing)

        ##### Compute general wing properties and things to be used in sectional analysis.
        sweep = wing.mean_sweep_angle()
        AR = wing.aspect_ratio()
        mach = op_point.mach()
        q = op_point.dynamic_pressure()
        CL_over_Cl = aerolib.CL_over_Cl(
            aspect_ratio=AR,
            mach=mach,
            sweep=sweep
        )
        oswalds_efficiency = aerolib.oswalds_efficiency(
            taper_ratio=wing.taper_ratio(),
            aspect_ratio=AR,
            sweep=sweep,
            fuselage_diameter_to_span_ratio=0  # an assumption
        )
        areas = wing.area(_sectional=True)
        aerodynamic_centers = wing.aerodynamic_center(_sectional=True)

        F_g = [0, 0, 0]
        M_g = [0, 0, 0]

        ##### Iterate through the wing sections.
        for sect_id in range(len(wing.xsecs) - 1):

            ##### Identify the wing cross sections adjacent to this wing section.
            xsec_a = wing.xsecs[sect_id]
            xsec_b = wing.xsecs[sect_id + 1]

            ##### When linearly interpolating, weight things by the relative chord.
            a_weight = xsec_a.chord / (xsec_a.chord + xsec_b.chord)
            b_weight = xsec_b.chord / (xsec_a.chord + xsec_b.chord)

            ##### Compute the local frame of this section, and put the z (normal) component into wind axes.
            xg_local, yg_local, zg_local = wing._compute_frame_of_section(sect_id)
            sect_aerodynamic_center = aerodynamic_centers[sect_id]

            sect_z_w = op_point.convert_axes(
                x_from=zg_local[0], y_from=zg_local[1], z_from=zg_local[2],
                from_axes="geometry",
                to_axes="wind",
            )

            ##### Compute the generalized angle of attack, so the geometric alpha that the wing section "sees".
            velocity_vector_b_from_freestream = op_point.convert_axes(
                x_from=-op_point.velocity, y_from=0, z_from=0,
                from_axes="wind",
                to_axes="body"
            )
            velocity_vector_b_from_rotation = np.cross(
                op_point.convert_axes(
                    sect_aerodynamic_center[0], sect_aerodynamic_center[1], sect_aerodynamic_center[2],
                    from_axes="geometry",
                    to_axes="body"
                ),
                [op_point.p, op_point.q, op_point.r],
                manual=True
            )
            velocity_vector_b = [
                velocity_vector_b_from_freestream[i] + velocity_vector_b_from_rotation[i]
                for i in range(3)
            ]
            velocity_mag_b = np.sqrt(sum([comp ** 2 for comp in velocity_vector_b]))
            velocity_dir_b = [
                velocity_vector_b[i] / velocity_mag_b
                for i in range(3)
            ]
            sect_z_b = op_point.convert_axes(
                x_from=zg_local[0], y_from=zg_local[1], z_from=zg_local[2],
                from_axes="geometry",
                to_axes="body",
            )
            vel_dot_normal = np.dot(velocity_dir_b, sect_z_b, manual=True)

            sect_alpha_generalized = 90 - np.arccosd(vel_dot_normal)

            def get_deflection(xsec):
                n_surfs = len(xsec.control_surfaces)
                if n_surfs == 0:
                    return 0
                elif n_surfs == 1:
                    surf = xsec.control_surfaces[0]
                    return surf.deflection
                else:
                    raise NotImplementedError(
                        "AeroBuildup currently cannot handle multiple control surfaces attached to a given WingXSec.")

            ##### Compute sectional lift at cross sections using lookup functions. Merge them linearly to get section CL.
            xsec_a_Cl_incompressible = xsec_a.airfoil.CL_function(
                alpha=sect_alpha_generalized,
                Re=op_point.reynolds(xsec_a.chord),
                mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
                deflection=get_deflection(xsec_a)
            )
            xsec_b_Cl_incompressible = xsec_b.airfoil.CL_function(
                alpha=sect_alpha_generalized,
                Re=op_point.reynolds(xsec_b.chord),
                mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
                deflection=get_deflection(xsec_b)
            )
            sect_CL = (
                              xsec_a_Cl_incompressible * a_weight +
                              xsec_b_Cl_incompressible * b_weight
                      ) * CL_over_Cl

            ##### Compute sectional drag at cross sections using lookup functions. Merge them linearly to get section CD.
            xsec_a_Cd_profile = xsec_a.airfoil.CD_function(
                alpha=sect_alpha_generalized,
                Re=op_point.reynolds(xsec_a.chord),
                mach=mach,
                deflection=get_deflection(xsec_a)
            )
            xsec_b_Cd_profile = xsec_b.airfoil.CD_function(
                alpha=sect_alpha_generalized,
                Re=op_point.reynolds(xsec_b.chord),
                mach=mach,
                deflection=get_deflection(xsec_b)
            )
            sect_CDp = (
                    xsec_a_Cd_profile * a_weight +
                    xsec_b_Cd_profile * b_weight
            )

            ##### Compute induced drag from local CL and full-wing properties (AR, e)
            sect_CDi = (
                    sect_CL ** 2 / (np.pi * AR * oswalds_efficiency)
            )

            ##### Total the drag.
            sect_CD = sect_CDp + sect_CDi

            ##### Go to dimensional quantities using the area.
            area = areas[sect_id]
            sect_L = q * area * sect_CL
            sect_D = q * area * sect_CD

            ##### Compute the direction of the lift by projecting the section's normal vector into the plane orthogonal to the freestream.
            sect_L_direction_w = (
                np.zeros_like(sect_z_w[0]),
                sect_z_w[1] / np.sqrt(sect_z_w[1] ** 2 + sect_z_w[2] ** 2),
                sect_z_w[2] / np.sqrt(sect_z_w[1] ** 2 + sect_z_w[2] ** 2)
            )
            sect_L_direction_g = op_point.convert_axes(
                *sect_L_direction_w, from_axes="wind", to_axes="geometry"
            )

            ##### Compute the direction of the drag by aligning the drag vector with the freestream vector.
            sect_D_direction_w = (-1, 0, 0)
            sect_D_direction_g = op_point.convert_axes(
                *sect_D_direction_w, from_axes="wind", to_axes="geometry"
            )

            ##### Compute the force vector in geometry axes.
            sect_F_g = [
                sect_L * sect_L_direction_g[i] + sect_D * sect_D_direction_g[i]
                for i in range(3)
            ]

            ##### Compute the moment vector in geometry axes.
            sect_M_g = np.cross(
                sect_aerodynamic_center,
                sect_F_g,
                manual=True
            )

            ##### Add section forces and moments to overall forces and moments
            F_g = [
                F_g[i] + sect_F_g[i]
                for i in range(3)
            ]
            M_g = [
                M_g[i] + sect_M_g[i]
                for i in range(3)
            ]

            ##### Treat symmetry
            if wing.symmetric:
                ##### Compute the local frame of this section, and put the z (normal) component into wind axes.

                sym_sect_aerodynamic_center = aerodynamic_centers[sect_id]
                sym_sect_aerodynamic_center[1] *= -1

                sym_sect_z_w = op_point.convert_axes(
                    x_from=zg_local[0], y_from=-zg_local[1], z_from=zg_local[2],
                    from_axes="geometry",
                    to_axes="wind",
                )

                ##### Compute the generalized angle of attack, so the geometric alpha that the wing section "sees".
                sym_velocity_vector_b_from_freestream = op_point.convert_axes(
                    x_from=-op_point.velocity, y_from=0, z_from=0,
                    from_axes="wind",
                    to_axes="body"
                )
                sym_velocity_vector_b_from_rotation = np.cross(
                    op_point.convert_axes(
                        sym_sect_aerodynamic_center[0], sym_sect_aerodynamic_center[1], sym_sect_aerodynamic_center[2],
                        from_axes="geometry",
                        to_axes="body"
                    ),
                    [op_point.p, op_point.q, op_point.r],
                    manual=True
                )
                sym_velocity_vector_b = [
                    sym_velocity_vector_b_from_freestream[i] + sym_velocity_vector_b_from_rotation[i]
                    for i in range(3)
                ]
                sym_velocity_mag_b = np.sqrt(sum([comp ** 2 for comp in sym_velocity_vector_b]))
                sym_velocity_dir_b = [
                    sym_velocity_vector_b[i] / sym_velocity_mag_b
                    for i in range(3)
                ]
                sym_sect_z_b = op_point.convert_axes(
                    x_from=zg_local[0], y_from=-zg_local[1], z_from=zg_local[2],
                    from_axes="geometry",
                    to_axes="body",
                )
                sym_vel_dot_normal = np.dot(sym_velocity_dir_b, sym_sect_z_b, manual=True)

                sym_sect_alpha_generalized = 90 - np.arccosd(sym_vel_dot_normal)

                def get_deflection(xsec):
                    n_surfs = len(xsec.control_surfaces)
                    if n_surfs == 0:
                        return 0
                    elif n_surfs == 1:
                        surf = xsec.control_surfaces[0]
                        return surf.deflection if surf.symmetric else -surf.deflection
                    else:
                        raise NotImplementedError(
                            "AeroBuildup currently cannot handle multiple control surfaces attached to a given WingXSec.")

                ##### Compute sectional lift at cross sections using lookup functions. Merge them linearly to get section CL.
                sym_xsec_a_Cl_incompressible = xsec_a.airfoil.CL_function(
                    alpha=sym_sect_alpha_generalized,
                    Re=op_point.reynolds(xsec_a.chord),
                    mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
                    deflection=get_deflection(xsec_a)
                )
                sym_xsec_b_Cl_incompressible = xsec_b.airfoil.CL_function(
                    alpha=sym_sect_alpha_generalized,
                    Re=op_point.reynolds(xsec_b.chord),
                    mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
                    deflection=get_deflection(xsec_b)
                )
                sym_sect_CL = (
                                      sym_xsec_a_Cl_incompressible * a_weight +
                                      sym_xsec_b_Cl_incompressible * b_weight
                              ) * CL_over_Cl

                ##### Compute sectional drag at cross sections using lookup functions. Merge them linearly to get section CD.
                sym_xsec_a_Cd_profile = xsec_a.airfoil.CD_function(
                    alpha=sym_sect_alpha_generalized,
                    Re=op_point.reynolds(xsec_a.chord),
                    mach=mach,
                    deflection=get_deflection(xsec_a)
                )
                sym_xsec_b_Cd_profile = xsec_b.airfoil.CD_function(
                    alpha=sym_sect_alpha_generalized,
                    Re=op_point.reynolds(xsec_b.chord),
                    mach=mach,
                    deflection=get_deflection(xsec_b)
                )
                sym_sect_CDp = (
                        sym_xsec_a_Cd_profile * a_weight +
                        sym_xsec_b_Cd_profile * b_weight
                )

                ##### Compute induced drag from local CL and full-wing properties (AR, e)
                sym_sect_CDi = (
                        sym_sect_CL ** 2 / (np.pi * AR * oswalds_efficiency)
                )

                ##### Total the drag.
                sym_sect_CD = sym_sect_CDp + sym_sect_CDi

                ##### Go to dimensional quantities using the area.
                area = areas[sect_id]
                sym_sect_L = q * area * sym_sect_CL
                sym_sect_D = q * area * sym_sect_CD

                ##### Compute the direction of the lift by projecting the section's normal vector into the plane orthogonal to the freestream.
                sym_sect_L_direction_w = (
                    np.zeros_like(sym_sect_z_w[0]),
                    sym_sect_z_w[1] / np.sqrt(sym_sect_z_w[1] ** 2 + sym_sect_z_w[2] ** 2),
                    sym_sect_z_w[2] / np.sqrt(sym_sect_z_w[1] ** 2 + sym_sect_z_w[2] ** 2)
                )
                sym_sect_L_direction_g = op_point.convert_axes(
                    *sym_sect_L_direction_w, from_axes="wind", to_axes="geometry"
                )

                ##### Compute the direction of the drag by aligning the drag vector with the freestream vector.
                sym_sect_D_direction_w = (-1, 0, 0)
                sym_sect_D_direction_g = op_point.convert_axes(
                    *sym_sect_D_direction_w, from_axes="wind", to_axes="geometry"
                )

                ##### Compute the force vector in geometry axes.
                sym_sect_F_g = [
                    sym_sect_L * sym_sect_L_direction_g[i] + sym_sect_D * sym_sect_D_direction_g[i]
                    for i in range(3)
                ]

                ##### Compute the moment vector in geometry axes.
                sym_sect_M_g = np.cross(
                    sym_sect_aerodynamic_center,
                    sym_sect_F_g,
                    manual=True
                )

                ##### Add section forces and moments to overall forces and moments
                F_g = [
                    F_g[i] + sym_sect_F_g[i]
                    for i in range(3)
                ]
                M_g = [
                    M_g[i] + sym_sect_M_g[i]
                    for i in range(3)
                ]

        ##### Convert F_g and M_g to body and wind axes for reporting.
        F_b = op_point.convert_axes(*F_g, from_axes="geometry", to_axes="body")
        F_w = op_point.convert_axes(*F_b, from_axes="body", to_axes="wind")
        M_b = op_point.convert_axes(*M_g, from_axes="geometry", to_axes="body")
        M_w = op_point.convert_axes(*M_b, from_axes="body", to_axes="wind")

        return {
            "F_g": F_g,
            "F_b": F_b,
            "F_w": F_w,
            "M_g": M_g,
            "M_b": M_b,
            "M_w": M_w,
            "L"  : -F_w[2],
            "Y"  : F_w[1],
            "D"  : -F_w[0],
            "l_b": M_b[0],
            "m_b": M_b[1],
            "n_b": M_b[2]
        }

    def fuselage_aerodynamics(self,
                              fuselage: Fuselage,
                              ):
        """
        Estimates the aerodynamic forces, moments, and derivatives on a fuselage in isolation.

        Assumes:
            * The fuselage is a body of revolution aligned with the x_b axis.
            * The angle between the nose and the freestream is less than 90 degrees.

        Moments are given with the reference at Fuselage [0, 0, 0].

        Uses methods from Jorgensen, Leland Howard. "Prediction of Static Aerodynamic Characteristics for Slender Bodies
        Alone and with Lifting Surfaces to Very High Angles of Attack". NASA TR R-474. 1977.

        Args:

            fuselage: A Fuselage object that you wish to analyze.

        Returns:

        """
        ##### Alias a few things for convenience
        op_point = self.op_point
        Re = op_point.reynolds(reference_length=fuselage.length())
        fuse_options = self.get_options(fuselage)

        ####### Reference quantities (Set these 1 here, just so we can follow Jorgensen syntax.)
        # Outputs of this function should be invariant of these quantities, if normalization has been done correctly.
        S_ref = 1  # m^2
        c_ref = 1  # m

        ####### Fuselage zero-lift drag estimation

        ### Forebody drag
        C_f_forebody = aerolib.Cf_flat_plate(
            Re_L=Re
        )

        ### Base Drag
        C_D_base = 0.029 / np.sqrt(C_f_forebody) * fuselage.area_base() / S_ref

        ### Skin friction drag
        C_D_skin = C_f_forebody * fuselage.area_wetted() / S_ref

        ### Wave drag
        if self.include_wave_drag:
            sears_haack_drag = transonic.sears_haack_drag_from_volume(
                volume=fuselage.volume(),
                length=fuselage.length()
            )
            C_D_wave = transonic.approximate_CD_wave(
                mach=op_point.mach(),
                mach_crit=critical_mach(
                    fineness_ratio_nose=fuse_options["nose_fineness_ratio"]
                ),
                CD_wave_at_fully_supersonic=fuse_options["E_wave_drag"] * sears_haack_drag,
            )
        else:
            C_D_wave = 0

        ### Total zero-lift drag
        C_D_zero_lift = C_D_skin + C_D_base + C_D_wave

        ####### Jorgensen model

        ### First, merge the alpha and beta into a single "generalized alpha", which represents the degrees between the fuselage axis and the freestream.
        x_w, y_w, z_w = op_point.convert_axes(
            1, 0, 0, from_axes="body", to_axes="wind"
        )
        generalized_alpha = np.arccosd(x_w / (1 + 1e-14))
        sin_generalized_alpha = np.sind(generalized_alpha)
        cos_generalized_alpha = x_w

        # ### Limit generalized alpha to -90 < alpha < 90, for now.
        # generalized_alpha = np.clip(generalized_alpha, -90, 90)
        # # TODO make the drag/moment functions not give negative results for alpha > 90.

        alpha_fractional_component = -z_w / np.sqrt(
            y_w ** 2 + z_w ** 2 + 1e-16)  # The fraction of any "generalized lift" to be in the direction of alpha
        beta_fractional_component = y_w / np.sqrt(
            y_w ** 2 + z_w ** 2 + 1e-16)  # The fraction of any "generalized lift" to be in the direction of beta

        ### Compute normal quantities
        ### Note the (N)ormal, (A)ligned coordinate system. (See Jorgensen for definitions.)
        # M_n = sin_generalized_alpha * op_point.mach()
        Re_n = sin_generalized_alpha * Re
        # V_n = sin_generalized_alpha * op_point.velocity
        q = op_point.dynamic_pressure()
        x_nose = fuselage.xsecs[0].xyz_c[0]
        x_m = 0 - x_nose
        x_c = fuselage.x_centroid_projected() - x_nose

        ##### Potential flow crossflow model
        C_N_p = (  # Normal force coefficient due to potential flow. (Jorgensen Eq. 2.12, part 1)
                fuselage.area_base() / S_ref * np.sind(2 * generalized_alpha) * np.cosd(generalized_alpha / 2)
        )
        C_m_p = (
                (
                        fuselage.volume() - fuselage.area_base() * (fuselage.length() - x_m)
                ) / (
                        S_ref * c_ref
                ) * np.sind(2 * generalized_alpha) * np.cosd(generalized_alpha / 2)
        )

        ##### Viscous crossflow model
        C_d_n = np.where(
            Re_n != 0,
            aerolib.Cd_cylinder(Re_D=Re_n),  # Replace with 1.20 from Jorgensen Table 1 if not working well
            0
        )
        eta = jorgensen_eta(fuselage.fineness_ratio())

        C_N_v = (  # Normal force coefficient due to viscous crossflow. (Jorgensen Eq. 2.12, part 2)
                eta * C_d_n * fuselage.area_projected() / S_ref * sin_generalized_alpha ** 2
        )
        C_m_v = (
                eta * C_d_n * fuselage.area_projected() / S_ref * (x_m - x_c) / c_ref * sin_generalized_alpha ** 2
        )

        ##### Total C_N model
        C_N = C_N_p + C_N_v
        C_m_generalized = C_m_p + C_m_v

        ##### Total C_A model
        C_A = C_D_zero_lift * cos_generalized_alpha * np.abs(cos_generalized_alpha)

        ##### Convert to lift, drag
        C_L_generalized = C_N * cos_generalized_alpha - C_A * sin_generalized_alpha
        C_D = C_N * sin_generalized_alpha + C_A * cos_generalized_alpha

        ### Set proper directions

        C_L = C_L_generalized * alpha_fractional_component
        C_Y = -C_L_generalized * beta_fractional_component
        C_l = 0
        C_m = C_m_generalized * alpha_fractional_component
        C_n = -C_m_generalized * beta_fractional_component

        ### Un-normalize
        L = C_L * q * S_ref
        Y = C_Y * q * S_ref
        D = C_D * q * S_ref
        l_w = C_l * q * S_ref * c_ref
        m_w = C_m * q * S_ref * c_ref
        n_w = C_n * q * S_ref * c_ref

        ### Convert to axes coordinates for reporting
        F_w = (
            -D,
            Y,
            -L
        )
        F_b = op_point.convert_axes(*F_w, from_axes="wind", to_axes="body")
        F_g = op_point.convert_axes(*F_b, from_axes="body", to_axes="geometry")
        M_w = (
            l_w,
            m_w,
            n_w,
        )
        M_b = op_point.convert_axes(*M_w, from_axes="wind", to_axes="body")
        M_g = op_point.convert_axes(*M_b, from_axes="body", to_axes="geometry")

        return {
            "F_g": F_g,
            "F_b": F_b,
            "F_w": F_w,
            "M_g": M_g,
            "M_b": M_b,
            "M_w": M_w,
            "L"  : -F_w[2],
            "Y"  : F_w[1],
            "D"  : -F_w[0],
            "l_b": M_b[0],
            "m_b": M_b[1],
            "n_b": M_b[2]
        }


if __name__ == '__main__':
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.conventional import airplane

    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(alpha=0, beta=1),
    ).run()

    from aerosandbox.tools.pretty_plots import plt, show_plot, contour, equal, set_ticks

    fig, ax = plt.subplots(2, 2)
    alpha = np.linspace(-10, 10, 1000)
    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(
            velocity=100,
            alpha=alpha,
            beta=0
        ),
    ).run()

    plt.sca(ax[0, 0])
    plt.plot(alpha, aero["CL"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_L$")
    set_ticks(5, 1, 0.5, 0.1)

    plt.sca(ax[0, 1])
    plt.plot(alpha, aero["CD"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_D$")
    set_ticks(5, 1, 0.05, 0.01)
    plt.ylim(bottom=0)

    plt.sca(ax[1, 0])
    plt.plot(alpha, aero["Cm"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_m$")
    set_ticks(5, 1, 0.5, 0.1)

    plt.sca(ax[1, 1])
    plt.plot(alpha, aero["CL"] / aero["CD"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_L/C_D$")
    set_ticks(5, 1, 10, 2)

    show_plot(
        "`asb.AeroBuildup` Aircraft Aerodynamics"
    )

    fig, ax = plt.subplots(figsize=(7, 6))
    Beta, Alpha = np.meshgrid(np.linspace(-90, 90, 200), np.linspace(-90, 90, 200))
    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(
            velocity=10,
            alpha=Alpha,
            beta=Beta
        ),
    ).run()
    contour(Beta, Alpha, aero["CL"], levels=30)
    equal()
    show_plot("AeroBuildup", r"$\beta$ [deg]", r"$\alpha$ [deg]")



from aerosandbox.common import ExplicitAnalysis
import aerosandbox.numpy as np
import subprocess
from pathlib import Path
from aerosandbox.geometry import Airplane, Wing, WingXSec, Fuselage, ControlSurface
from aerosandbox.performance import OperatingPoint
from typing import Union, List, Dict, Any
import copy
import tempfile
import warnings


class AVL(ExplicitAnalysis):
    """

    An interface to AVL, a 3D vortex lattice aerodynamics code developed by Mark Drela at MIT.

    Requires AVL to be on your computer; AVL is available here: https://web.mit.edu/drela/Public/web/avl/

    It is recommended (but not required) that you add AVL to your system PATH environment variable such that it can
    be called with the command `avl`. If this is not the case, you need to specify the path to your AVL
    executable using the `avl_command` argument of the constructor.

    Usage example:

        >>> avl = asb.AVL(
        >>>     airplane=my_airplane,
        >>>     op_point=asb.OperatingPoint(
        >>>         velocity=100, # m/s
        >>>         alpha=5, # deg
        >>>         beta=4, # deg
        >>>         p=0.01, # rad/sec
        >>>         q=0.02, # rad/sec
        >>>         r=0.03, # rad/sec
        >>>     )
        >>> )
        >>> outputs = avl.run()

    """
    default_analysis_specific_options = {
        Airplane: dict(
            profile_drag_coefficient=0
        ),
        Wing    : dict(
            wing_level_spanwise_spacing=True,
            spanwise_resolution=12,
            spanwise_spacing="cosine",
            chordwise_resolution=12,
            chordwise_spacing="cosine",
            component=None,  # type: int
            no_wake=False,
            no_alpha_beta=False,
            no_load=False,
            drag_polar=dict(
                CL1=0,
                CD1=0,
                CL2=0,
                CD2=0,
                CL3=0,
                CD3=0,
            ),
        ),
        WingXSec: dict(
            spanwise_resolution=12,
            spanwise_spacing="cosine",
            cl_alpha_factor=None, # type: float
            drag_polar=dict(
                CL1=0,
                CD1=0,
                CL2=0,
                CD2=0,
                CL3=0,
                CD3=0,
            )
        ),
        Fuselage: dict(
            panel_resolution=24,
            panel_spacing="cosine"
        )
    }

    AVL_spacing_parameters = {
        "uniform": 0,
        "cosine" : 1,
        "sine"   : 2,
        "-sine"  : -2,
        "equal"  : 0,  # "uniform" is preferred
    }

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint = None,
                 avl_command: str = "avl",
                 verbose: bool = False,
                 working_directory: str = None,
                 ground_effect: bool = False,
                 ground_effect_height: float = 0
                 ):
        """
        Interface to AVL.

        Args:

            airplane: The airplane object you wish to analyze.

            op_point: The operating point you wish to analyze at.

            avl_command: The command-line argument to call AVL.

                * If AVL is on your system PATH, then you can just leave this as "avl".

                * If AVL is not on your system PATH, thjen you should provide a filepath to the AVL executable.

                Note that AVL is not on your PATH by default. To tell if AVL is on your system PATH, open up a
                terminal and type "avl".

                    * If the AVL menu appears, it's on your PATH.

                    * If you get something like "'avl' is not recognized as an internal or external command..." or
                    "Command 'avl' not found, did you mean...", then it is not on your PATH and you'll need to
                    specify the location of your AVL executable as a string.

                To add AVL to your path, modify your system's environment variables. (Google how to do this for your OS.)

            verbose:

            working_directory:
        """
        ### Set defaults
        if op_point is None:
            op_point = OperatingPoint()

        ### Initialize
        self.airplane = airplane
        self.op_point = op_point
        self.avl_command = avl_command
        self.verbose = verbose
        self.working_directory = working_directory
        self.ground_effect = ground_effect
        self.ground_effect_height = ground_effect_height

    def run(self) -> Dict:
        return self._run_avl()

    def _default_keystroke_file_contents(self) -> List[str]:

        run_file_contents = []

        # Disable graphics
        run_file_contents += [
            "plop",
            "g",
            "",
        ]

        # Enter oper mode
        run_file_contents += [
            "oper",
        ]

        # Set parameters
        run_file_contents += [
            "m",
            f"mn {self.op_point.mach()}",
            f"v {self.op_point.velocity}",
            f"d {self.op_point.atmosphere.density()}",
            "g 9.81",
            ""
        ]

        # Set analysis state
        p_bar = self.op_point.p * self.airplane.b_ref / (2 * self.op_point.velocity)
        q_bar = self.op_point.q * self.airplane.c_ref / (2 * self.op_point.velocity)
        r_bar = self.op_point.r * self.airplane.b_ref / (2 * self.op_point.velocity)

        run_file_contents += [
            f"a a {self.op_point.alpha}",
            f"b b {self.op_point.beta}",
            f"r r {p_bar}",
            f"p p {q_bar}",
            f"y y {r_bar}"
        ]

        # Set control surface deflections
        run_file_contents += [
            f"d1 d1 1"
        ]

        return run_file_contents

    def _run_avl(self,
                 run_command: str = None,
                 ) -> Dict[str, np.ndarray]:
        """
        Private function to run AVL.

        Args: run_command: A string with any AVL keystroke inputs that you'd like. By default, you start off within the OPER
        menu. All of the inputs indicated in the constructor have been set already, but you can override them here (
        for this run only) if you want.

        Returns: A dictionary containing all of your results.

        """
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            ### Alternatively, work in another directory:
            if self.working_directory is not None:
                directory = Path(self.working_directory)  # For debugging

            # Designate an intermediate file for file I/O
            output_filename = "output.txt"
            with open(directory / output_filename, "w+") as f:
                pass

            # Handle the airplane file
            airplane_file = "airplane.avl"
            self.write_avl(directory / airplane_file)

            # Handle the run file
            keystroke_file_contents = self._default_keystroke_file_contents()
            if run_command is not None:
                keystroke_file_contents += [run_command]
            keystroke_file_contents += [
                "x",
                "st",
                f"{output_filename}",
                "o",
                "",
                "",
                "quit"
            ]
            keystroke_file = "keystroke_file.txt"
            with open(directory / keystroke_file, "w+") as f:
                f.write(
                    "\n".join(keystroke_file_contents)
                )

            command = f'{self.avl_command} {airplane_file} < {keystroke_file}'

            ### Execute
            subprocess.call(
                command,
                shell=True,
                cwd=directory,
                stdout=None if self.verbose else subprocess.DEVNULL
            )

            ##### Parse the output file
            # Read the file
            with open(directory / output_filename, "r") as f:
                output_data = f.read()

            # Trim off the first few lines that contain name, # of panels, etc.
            output_data = "\n".join(output_data.split("\n")[8:])

            ### Iterate through the string to find all the keys and corresponding numeric values, based on where "=" appears.
            keys = []
            values = []
            index = output_data.find("=")

            while index != -1:

                # All keys contain no spaces except for "Clb Cnr / Clr Cnb" (spiral stability parameter)
                potential_keys = output_data[:index].split()
                if potential_keys[-1] == "Cnb" and len(potential_keys) >= 5:
                    potential_key = " ".join(potential_keys[-5:])
                    if potential_key == "Clb Cnr / Clr Cnb":
                        key = potential_key
                    else:
                        key = potential_keys[-1]
                else:
                    key = potential_keys[-1]

                keys.append(key)
                output_data = output_data[index + 1:]
                number = output_data[:12].split("\n")[0]
                try:
                    number = float(number)
                except:
                    number = np.nan
                values.append(number)

                index = output_data.find("=")

            # Make key names consistent with AeroSandbox notation
            keys_lowerize = ["Alpha", "Beta", "Mach"]
            keys = [key.lower() if key in keys_lowerize else key for key in keys]
            keys = [key.replace("tot", "") for key in keys]

            if len(values) < 56:  # Sometimes the spiral mode term is inexplicably not displayed by AVL
                raise RuntimeError(
                    "AVL could not run for some reason!\n"
                    "Investigate by turning on the `verbose` flag and looking at the output.\n"
                    "(Common culprit: angular rates too high.)"
                )

            res = {
                k: v
                for k, v in zip(
                    keys, values
                )
            }

            ##### Add a few more outputs for ease of use
            res["p"] = res["pb/2V"] * (2 * self.op_point.velocity / self.airplane.b_ref)
            res["q"] = res["qc/2V"] * (2 * self.op_point.velocity / self.airplane.c_ref)
            res["r"] = res["rb/2V"] * (2 * self.op_point.velocity / self.airplane.b_ref)

            return res

    def write_avl(self,
                  filepath: Union[Path, str] = None,
                  ) -> None:
        """
        Writes a .avl file corresponding to this airplane to a filepath.

        For use with the AVL vortex-lattice-method aerodynamics analysis tool by Mark Drela at MIT.
        AVL is available here: https://web.mit.edu/drela/Public/web/avl/

        Args:
            filepath: filepath (including the filename and .avl extension) [string]
                If None, this function returns the .avl file as a string.

        Returns: None

        """

        def clean(s):
            """
            Removes leading and trailing whitespace from each line of a multi-line string.
            """
            return "\n".join([line.strip() for line in s.split("\n")])

        # def print_control_surface(control_surface: ControlSurface,
        #                           options_for_analysis: Dict[type, Dict[str, Any]]
        #                           ) -> str:
        #
        #     if control_surface.trailing_edge:
        #         hinge_point = control_surface.hinge_point
        #     else:
        #         hinge_point = -control_surface.hinge_point  # leading edge surfaces are defined with negative hinge points in AVL
        #
        #     if options_for_analysis["duplication_factor"] is None:
        #         duplication_factor = 1.0 if control_surface.symmetric else -1.0
        #     else:
        #         duplication_factor = options_for_analysis["duplication_factor"]
        #
        #     string = clean(f"""
        #     CONTROL
        #     #Cname Cgain Xhinge HingeVec SgnDup
        #     control{control_surface.id} {options_for_analysis['gain']} {hinge_point} {' '.join(str(x) for x in options_for_analysis['hinge_vector'])} {duplication_factor}
        #     """)
        #
        #     return string
        #
        # def print_control_surface_deprecated(xsec: WingXSec,
        #                                      id: int) -> str:
        #
        #     sign_duplication = 1.0 if xsec.control_surface_is_symmetric else -1.0
        #     string = clean(f"""
        #     CONTROL
        #     #Cname Cgain Xhinge HingeVec SgnDup
        #     control{id} 1.0 {xsec.control_surface_hinge_point} 0.0 0.0 0.0 {sign_duplication}
        #     """)
        #
        #     return string

        airplane = self.airplane

        avl_file = ""

        airplane_options = self.get_options(airplane)

        avl_file += clean(f"""\
        {airplane.name}
        #Mach
        0
        #IYsym   IZsym   Zsym
         0       {1 if self.ground_effect else 0}   {self.ground_effect_height}
        #Sref    Cref    Bref
        {airplane.s_ref} {airplane.c_ref} {airplane.b_ref}
        #Xref    Yref    Zref
        {airplane.xyz_ref[0]} {airplane.xyz_ref[1]} {airplane.xyz_ref[2]}
        # CDp
        {airplane_options["profile_drag_coefficient"]}
        """)

        control_surface_counter = 1
        for wing in airplane.wings:

            wing_options = self.get_options(wing)

            spacing_line = f"{wing_options['chordwise_resolution']}   {self.AVL_spacing_parameters[wing_options['chordwise_spacing']]}"
            if wing_options["wing_level_spanwise_spacing"]:
                spacing_line += f"   {wing_options['spanwise_resolution']}   {self.AVL_spacing_parameters[wing_options['spanwise_spacing']]}"

            avl_file += clean(f"""\
            #{"=" * 50}
            SURFACE
            {wing.name}
            #Nchordwise  Cspace  [Nspanwise   Sspace]
            {spacing_line}
            
            """)

            if wing_options["component"] is not None:
                avl_file += clean(f"""\
                COMPONENT
                {wing_options['component']}
                    
                """)

            if wing.symmetric:
                avl_file += clean(f"""\
                YDUPLICATE
                0
                    
                """)

            if wing_options["no_wake"]:
                avl_file += clean(f"""\
                NOWAKE
                
                """)

            if wing_options["no_alpha_beta"]:
                avl_file += clean(f"""\
                NOALBE
                
                """)

            if wing_options["no_load"]:
                avl_file += clean(f"""\
                NOLOAD
                
                """)

            polar = wing_options["drag_polar"]
            avl_file += clean(f"""\
            CDCL
            #CL1  CD1  CL2  CD2  CL3  CD3
            {polar["CL1"]} {polar["CD1"]} {polar["CL2"]} {polar["CD2"]} {polar["CL3"]} {polar["CD3"]}
            
            """)

            ### Build up a buffer of the control surface strings to write to each section
            control_surface_commands: List[List[str]] = [
                []
                for _ in wing.xsecs
            ]
            for i, xsec in enumerate(wing.xsecs[:-1]):
                for surf in xsec.control_surfaces:
                    xhinge = surf.hinge_point if surf.trailing_edge else -surf.hinge_point
                    xyz_hinge_vector = wing._compute_frame_of_section(i)[1]
                    sign_dup = 1 if surf.symmetric else -1

                    command = clean(f"""\
                        CONTROL
                        #name, gain, Xhinge, XYZhvec, SgnDup
                        all_deflections {surf.deflection} {xhinge} {xyz_hinge_vector[0]} {xyz_hinge_vector[1]} {xyz_hinge_vector[2]} {sign_dup}
                        """)

                    control_surface_commands[i].append(command)
                    control_surface_commands[i + 1].append(command)

            ### Write the commands for each wing section
            for i, xsec in enumerate(wing.xsecs):

                xsec_options = self.get_options(xsec)

                xsec_def_line = f"{xsec.xyz_le[0]} {xsec.xyz_le[1]} {xsec.xyz_le[2]} {xsec.chord} {xsec.twist}"
                if not wing_options["wing_level_spanwise_spacing"]:
                    xsec_def_line += f"   {xsec_options['spanwise_resolution']}   {self.AVL_spacing_parameters[xsec_options['spanwise_spacing']]}"

                if xsec_options["cl_alpha_factor"] is None:
                    claf_line = f"{1 + 0.77 * xsec.airfoil.max_thickness()}  # Computed using rule from avl_doc.txt"
                else:
                    claf_line = f"{xsec_options['cl_alpha_factor']}"

                avl_file += clean(f"""\
                #{"-" * 50}
                SECTION
                #Xle    Yle    Zle     Chord   Ainc  [Nspanwise   Sspace]
                {xsec_def_line}
                
                AIRFOIL
                {xsec.airfoil.repanel(50).write_dat(filepath=None, include_name=False)}
                
                CLAF
                {claf_line}
                
                """)

                polar = xsec_options["drag_polar"]
                avl_file += clean(f"""\
                CDCL
                #CL1  CD1  CL2  CD2  CL3  CD3
                {polar["CL1"]} {polar["CD1"]} {polar["CL2"]} {polar["CD2"]} {polar["CL3"]} {polar["CD3"]}
                
                """)

                for control_surface_command in control_surface_commands[i]:
                    avl_file += control_surface_command

        filepath = Path(filepath)
        for i, fuse in enumerate(airplane.fuselages):
            fuse_filepath = Path(str(filepath) + f".fuse{i}")
            self.write_avl_bfile(
                fuselage=fuse,
                filepath=fuse_filepath
            )
            fuse_options = self.get_options(fuse)

            avl_file += clean(f"""\
            #{"=" * 50}
            BODY
            {fuse.name}
            {fuse_options['panel_resolution']} {self.AVL_spacing_parameters[fuse_options['panel_spacing']]}
            
            BFIL
            {fuse_filepath}
            
            """)

        if filepath is not None:
            with open(filepath, "w+") as f:
                f.write(avl_file)

    @staticmethod
    def write_avl_bfile(fuselage,
                        filepath: Union[Path, str] = None,
                        include_name: bool = True,
                        ) -> str:
        """
        Writes an AVL-compatible BFILE corresponding to this fuselage to a filepath.

        For use with the AVL vortex-lattice-method aerodynamics analysis tool by Mark Drela at MIT.
        AVL is available here: https://web.mit.edu/drela/Public/web/avl/

        Args:
            filepath: filepath (including the filename and .avl extension) [string]
                If None, this function returns the would-be file contents as a string.

            include_name: Should the name of the fuselage be included in the .dat file? (This should be True for use with AVL.)

        Returns:

        """
        filepath = Path(filepath)

        contents = []

        if include_name:
            contents += [fuselage.name]

        contents += [
                        f"{xyz_c[0]} {xyz_c[2] + r}"
                        for xyz_c, r in zip(
                [xsec.xyz_c for xsec in fuselage.xsecs][::-1],
                [xsec.radius for xsec in fuselage.xsecs][::-1]
            )
                    ] + [
                        f"{xyz_c[0]} {xyz_c[2] - r}"
                        for xyz_c, r in zip(
                [xsec.xyz_c for xsec in fuselage.xsecs][1:],
                [xsec.radius for xsec in fuselage.xsecs][1:]
            )
                    ]

        string = "\n".join(contents)

        if filepath is not None:
            with open(filepath, "w+") as f:
                f.write(string)

        return string


if __name__ == '__main__':

    ### Import Vanilla Airplane
    import aerosandbox as asb

    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.vanilla import airplane as vanilla

    vanilla.analysis_specific_options[AVL] = dict(
        profile_drag_coefficient=0.1
    )
    vanilla.wings[0].xsecs[0].control_surfaces.append(
        ControlSurface(
            name="Flap",
            trailing_edge=True,
            hinge_point=0.75,
            symmetric=True,
            deflection=10
        )
    )
    ### Do the AVL run
    avl = AVL(
        airplane=vanilla,
        op_point=OperatingPoint(
            atmosphere=asb.Atmosphere(altitude=0),
            velocity=1,
            alpha=0.433476,
            beta=0,
            p=0,
            q=0,
            r=0,
        ),
        # working_directory=str(Path.home() / "Downloads" / "avl_test"),
        # verbose=True
    )

    res = avl.run()

    for k, v in res.items():
        print(f"{str(k).rjust(10)} : {v}")



import inspect
from typing import List
from pathlib import Path
from aerosandbox.tools.string_formatting import has_balanced_parentheses


def get_caller_source_location(
        stacklevel: int = 1,
) -> (Path, int):
    ### Go up `stacklevel` frames from the current one to get to the caller frame.
    frame = inspect.currentframe()
    for _ in range(stacklevel):
        frame = frame.f_back

    ### Extract the frame info (an `inspect.Traceback` type) from the caller frame
    frame_info: inspect.Traceback = inspect.getframeinfo(frame)

    filename = Path(frame_info.filename)
    lineno = frame_info.lineno

    return filename, lineno


def get_caller_source_code(
        stacklevel: int = 1,
        strip_lines: bool = False
) -> str:
    """
    Gets the source code of wherever this function is called.

    You can get source code from further up the call stack by using the `stacklevel` argument.

    Args:

        stacklevel: Choose the level of the stack that you want to retrieve source code at. Higher integers will get
        you higher (i.e., more end-user-facing) in the stack. Same behaviour as the `stacklevel` argument in
        warnings.warn().

        strip_lines: A boolean flag about whether or not to strip leading and trailing whitespace off each line of a
        multi-line function call. See the built-in string method `str.strip()` for behaviour.

    Returns: The source code of the call, as a string. Might be a multi-line string (i.e., contains '\n' characters)
    if the call is multi-line.

    """
    ### Go up `stacklevel` frames from the current one to get to the caller frame.
    frame = inspect.currentframe()
    for _ in range(stacklevel):
        frame = frame.f_back

    ### Extract the frame info (an `inspect.Traceback` type) from the caller frame
    frame_info: inspect.Traceback = inspect.getframeinfo(frame)

    ### If Python's auto-extracted "code context" is a compete statement, then you're done here.
    code_context = "".join(frame_info.code_context)
    if has_balanced_parentheses(code_context):
        return code_context

    ### Initialize the caller source lines, which is a list of strings that contain the source for the call.
    caller_source_lines: List[str] = []

    ### Read the source lines of code surrounding the call
    try:
        with open(frame_info.filename, "r") as f:  # Read the file containing the call
            for _ in range(frame_info.lineno - 1):  # Skip the first N lines of code, until you get to the call
                f.readline()  # Unfortunately there's no way around this, since you need to find the "\n" encodings in the file

            parenthesis_level = 0  # Track the number of "(" and ")" characters, so you know when the function call is complete

            def add_line() -> None:
                """
                Adds the subsequent line to the caller source lines (`caller_source_lines`). In-place.
                """
                line = f.readline()

                if strip_lines:
                    line = line.strip()

                nonlocal parenthesis_level
                for char in line:
                    if char == "(":
                        parenthesis_level += 1
                    elif char == ")":
                        parenthesis_level -= 1
                caller_source_lines.append(line)

            ### Get the first line, which is always part of the function call, and includes the opening parenthesis
            add_line()

            ### Do subsequent lines
            while parenthesis_level > 0:
                add_line()
    except OSError as e:
        raise FileNotFoundError(
            "\n".join([
                "Couldn't retrieve source code at this stack level, because the source code file couldn't be opened for some reason.",
                "One common possible reason is that you're referring to an IPython console with a multi-line statement."
            ])
        )

    caller_source = "\n".join(caller_source_lines)

    return caller_source


def get_function_argument_names_from_source_code(source_code: str) -> List[str]:
    """
    Gets the names of the function arguments found in a particular line of source code.

    Specifically, it retrieves the names of the arguments in the first function call found in the source code string.

    If the source code line is an assignment statement, only the right-hand-side of the line is analyzed.

    Also, removes all line breaks ('\n').

    Examples function inputs and outputs:

        "f(a, b)"               -> ['a', 'b']
        "f(a,b)"                -> ['a', 'b']
        "f(\na,\nb)"            -> ['a', 'b']
        "g = f(a, b)"           -> ['a', 'b']
        "g.h = f(a, b)"         -> ['a', 'b']
        "g.h() = f(a, b)"       -> ['a', 'b']
        "g.h(i=j) = f(a, b)"    -> ['a', 'b']
        "f(a, b) + g(h)"        -> ['a', 'b']
        "f(a: int, b: float())" -> ['a', 'b']
        "f(a, b).g(c, d)"       -> ['a', 'b']
        "f(a(b), c)"            -> ['a(b)', 'c']
        "f(a(b,c), d)"          -> ['a(b,c)', 'd']
        "f({a:b}, c)"           -> ['{a:b}', 'c']
        "f(a[b], c)"            -> ['a[b]', 'c']
        "f({a:b, c:d}, e)"      -> ['{a:b,c:d}', 'e']
        "f({a:b,\nc:d}, e)"     -> ['{a:b,c:d}', 'e']
        "f(dict(a=b,c=d), e)"   -> ['dict(a=b,c=d)', 'e']
        "f(a=1, b=2)"           -> ['a=1', 'b=2']
        "f()"                   -> ['']
        "3 + 5"                 -> raises ValueError
        ""                      -> raises ValueError

    Args:
        source_code: A line of Python source code that includes a function call. Can be a multi-line piece of source code (e.g., includes '\n').

    Returns: A list of strings containing all of the function arguments. If keyword arguments are found, includes both the key and the value, as-written.

    """

    assignment_equals_index = 0

    parenthesis_level = 0
    for i, char in enumerate(source_code):
        if char == "(":
            parenthesis_level += 1
        elif char == ")":
            parenthesis_level -= 1
        elif char == "=" and parenthesis_level == 0:
            assignment_equals_index = i + 1
            break

    source_code_rhs = source_code[assignment_equals_index:]

    source_code_rhs = source_code_rhs.replace("\n", "")

    parenthesis_level = 0
    braces_level = 0
    for i, char in enumerate(source_code_rhs):
        if char == "(":
            parenthesis_level += 1
            break

    if parenthesis_level == 0:
        raise ValueError("No function call was found in the source code provided!")

    arg_names: List[str] = []
    current_arg = ""
    in_type_hinting_block = False

    while parenthesis_level != 0:
        i += 1
        char = source_code_rhs[i]

        if char == "(":
            parenthesis_level += 1
        elif char == ")":
            parenthesis_level -= 1
        elif char == "{":
            braces_level += 1
        elif char == "}":
            braces_level -= 1

        if char == "," and parenthesis_level == 1 and braces_level == 0:
            arg_names.append(current_arg)
            current_arg = ""
            in_type_hinting_block = False
        elif char == ":" and parenthesis_level == 1 and braces_level == 0:
            in_type_hinting_block = True
        elif char == " ":
            pass
        elif parenthesis_level >= 1 and not in_type_hinting_block:
            current_arg += char

    arg_names.append(current_arg)

    def clean(s: str) -> str:
        return s.strip()

    arg_names = [
        clean(arg) for arg in arg_names
    ]

    return arg_names


if __name__ == '__main__':
    def dashes():
        """A quick macro for drawing some dashes, to make the terminal output clearer to distinguish."""
        print("\n" + "-" * 50 + "\n")


    dashes()

    print("Caller location:\n", get_caller_source_location(stacklevel=1))

    dashes()

    print("Caller source code:\n", get_caller_source_code(stacklevel=1))

    dashes()


    def my_func():
        print(
            get_caller_source_code(
                stacklevel=2
            )
        )


    print("Caller source code of a function call:")

    if_you_can_see_this_it_works = my_func()

    dashes()

    print("Arguments of f(a, b):")

    print(
        get_function_argument_names_from_source_code("f(a, b)")
    )



import aerosandbox.numpy as np
from typing import Union, List


def linear_hermite_patch(
        x: Union[float, np.ndarray],
        x_a: float,
        x_b: float,
        f_a: float,
        f_b: float,
):
    return (x - x_a) * (f_b - f_a) / (x_b - x_a) + f_a


def cubic_hermite_patch(
        x: Union[float, np.ndarray],
        x_a: float,
        x_b: float,
        f_a: float,
        f_b: float,
        dfdx_a: float,
        dfdx_b: float,
):
    dx = x_b - x_a
    t = (x - x_a) / dx  # Nondimensional distance along the patch
    return (
            (t ** 3) * (1 * f_b) +
            (t ** 2 * (1 - t)) * (3 * f_b - 1 * dfdx_b * dx) +
            (t * (1 - t) ** 2) * (3 * f_a + 1 * dfdx_a * dx) +
            ((1 - t) ** 3) * (1 * f_a)
    )


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    x = np.linspace(0.8, 1)
    plt.plot(
        x,
        cubic_hermite_patch(
            x,
            x_a=x.min(),
            x_b=x.max(),
            f_a=0.002,
            f_b=0.5,
            dfdx_a=0.1,
            dfdx_b=10
        )
    )

    p.equal()
    p.show_plot()




def critical_mach(fineness_ratio_nose: float) -> float:
    """
    Returns the transonic critical Mach number for a streamlined fuselage.

    Fitted to data from Raymer "Aircraft Design: A Conceptual Approach" 2nd Ed., Fig. 12.28.
    See figure + study + fit in: /studies/FuselageCriticalMach/

    Args:

        fineness_ratio_nose: The fineness ratio of the nose section of the fuselage.

            Specifically, fineness_ratio_nose = 2 * L_n / d, where:

                * L_n is the length from the nose to the longitudinal location at which the fuselage cross section
                becomes essentially constant, and:

                * d is the body diameter at that location.

    Returns: The critical Mach number

    """
    p = {
        'a': 11.087202397070559,
        'b': 13.469755774708842,
        'c': 4.034476257077558
    }

    mach_dd = 1 - (p["a"] / (2 * fineness_ratio_nose + p["b"])) ** p["c"]

    ### The following approximate relation is derived in W.H. Mason, "Configuration Aerodynamics", Chapter 7. Transonic Aerodynamics of Airfoils and Wings.
    ### Equation 7-8 on Page 7-19.
    ### This is in turn based on Lock's proposed empirically-derived shape of the drag rise, from Hilton, W.F., High Speed Aerodynamics, Longmans, Green & Co., London, 1952, pp. 47-49
    mach_crit = mach_dd - (0.1 / 80) ** (1 / 3)

    return mach_crit


def jorgensen_eta(fineness_ratio: float) -> float:
    """
    A fit for the eta parameter (crossflow lift multiplier) of a fuselage, as described in:

    Jorgensen, Leland Howard. "Prediction of Static Aerodynamic Characteristics for Slender Bodies
    Alone and with Lifting Surfaces to Very High Angles of Attack". NASA TR R-474. 1977.

    Fits performed in /studies/JorgensenEtaFitting/

    Args:
        fineness_ratio: The fineness ratio of the fuselage. (length / diameter)

    Returns: An estimate of eta.

    """
    x = fineness_ratio
    p = {
        '1scl': 23.009059965179222,
        '1cen': -122.76900250914575,
        '2scl': 13.006453125841258,
        '2cen': -24.367562906887436
    }
    return 1 - p["1scl"] / (x - p["1cen"]) - (p["2scl"] / (x - p["2cen"])) ** 2


import aerosandbox.numpy as np


def solar_flux_outside_atmosphere_normal(day_of_year):
    """
    Normal solar flux at the top of the atmosphere (variation due to orbital eccentricity)
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :return: Solar flux [W/m^2]
    """
    # Space effects
    # # Source: https://www.itacanet.org/the-sun-as-a-source-of-energy/part-2-solar-energy-reaching-the-earths-surface/#2.1.-The-Solar-Constant
    # solar_flux_outside_atmosphere_normal = 1367 * (1 + 0.034 * cas.cos(2 * cas.pi * (day_of_year / 365.25)))  # W/m^2; variation due to orbital eccentricity
    # Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/solar-radiation-outside-the-earths-atmosphere
    return 1366 * (
            1 + 0.033 * np.cosd(360 * (day_of_year - 2) / 365))  # W/m^2; variation due to orbital eccentricity


def declination_angle(day_of_year):
    """
    Declination angle, in degrees, as a func. of day of year. (Seasonality)
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :return: Declination angle [deg]
    """
    # Declination (seasonality)
    # Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/declination-angle
    return -23.45 * np.cosd(360 / 365 * (day_of_year + 10))  # in degrees


def solar_elevation_angle(latitude, day_of_year, time):
    """
    Elevation angle of the sun [degrees] for a local observer.

    :param latitude: Latitude [degrees]

    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)

    :param time: Time after local solar noon [seconds]

    :return: Solar elevation angle [degrees] (angle between horizon and sun). Returns negative values if the sun is
    below the horizon.
    """

    # Solar elevation angle (including seasonality, latitude, and time of day)
    # Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/elevation-angle
    declination = declination_angle(day_of_year)

    solar_elevation_angle = np.arcsind(
        np.sind(declination) * np.sind(latitude) +
        np.cosd(declination) * np.cosd(latitude) * np.cosd(time / 86400 * 360)
    )  # in degrees
    return solar_elevation_angle


def solar_azimuth_angle(latitude, day_of_year, time):
    """
    Azimuth angle of the sun [degrees] for a local observer.
    :param latitude: Latitude [degrees]
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :param time: Time after local solar noon [seconds]
    :return: Solar azimuth angle [degrees] (the compass direction from which the sunlight is coming).
    """

    # Solar azimuth angle (including seasonality, latitude, and time of day)
    # Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/azimuth-angle
    declination = declination_angle(day_of_year)
    sdec = np.sind(declination)
    cdec = np.cosd(declination)
    slat = np.sind(latitude)
    clat = np.cosd(latitude)
    ctime = np.cosd(time / 86400 * 360)

    elevation = solar_elevation_angle(latitude, day_of_year, time)
    cele = np.cosd(elevation)

    cos_azimuth = (sdec * clat - cdec * slat * ctime) / cele
    cos_azimuth = np.clip(cos_azimuth, -1, 1)

    azimuth_raw = np.arccosd(cos_azimuth)

    is_solar_morning = np.mod(time, 86400) > 43200

    solar_azimuth_angle = np.where(
        is_solar_morning,
        azimuth_raw,
        360 - azimuth_raw
    )

    return solar_azimuth_angle


def incidence_angle_function(
        latitude: float,
        day_of_year: float,
        time: float,
        panel_azimuth_angle: float = 0,
        panel_tilt_angle: float = 0,
        scattering: bool = True,
):
    """
    This website will be useful for accounting for direction of the vertical surface
    https://www.pveducation.org/pvcdrom/properties-of-sunlight/arbitrary-orientation-and-tilt
    :param latitude: Latitude [degrees]
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :param time: Time since (local) solar noon [seconds]
    :param panel_azimuth_angle: The azimuth angle of the panel normal, in degrees. (0 degrees if pointing North and 90 if East)
    :param panel_tilt_angle: The angle between the panel normal and vertical, in degrees. (0 if horizontal and 90 if vertical)
    :param scattering: Boolean: include scattering effects at very low angles?

    :returns
    illumination_factor: Fraction of solar insolation received, relative to what it would get if it were perfectly oriented to the sun.
    """
    solar_elevation = solar_elevation_angle(latitude, day_of_year, time)
    solar_azimuth = solar_azimuth_angle(latitude, day_of_year, time)
    cosine_factor = (
            np.cosd(solar_elevation) *
            np.sind(panel_tilt_angle) *
            np.cosd(panel_azimuth_angle - solar_azimuth)
            + np.sind(solar_elevation) * np.cosd(panel_tilt_angle)
    )
    if scattering:
        illumination_factor = cosine_factor * scattering_factor(solar_elevation)
    else:
        illumination_factor = cosine_factor

    illumination_factor = np.fmax(illumination_factor, 0)
    illumination_factor = np.where(
        solar_elevation < 0,
        0,
        illumination_factor
    )
    return illumination_factor


def scattering_factor(elevation_angle):
    """
    Calculates a scattering factor (a factor that gives losses due to atmospheric scattering at low elevation angles).
    Source: AeroSandbox/studies/SolarPanelScattering
    :param elevation_angle: Angle between the horizon and the sun [degrees]
    :return: Fraction of the light that is not lost to scattering.
    # TODO figure out if scattering is w.r.t. sun elevation or w.r.t. angle between panel and sun
    """
    elevation_angle = np.clip(elevation_angle, 0, 90)
    theta = 90 - elevation_angle  # Angle between panel normal and the sun, in degrees

    # # Model 1
    # c = (
    #     0.27891510500505767300438719757949,
    #     -0.015994330894744987481281839336589,
    #     -19.707332432605799255043166340329,
    #     -0.66260979582573353852126274432521
    # )
    # factor = c[0] + c[3] * theta_rad + cas.exp(
    #     c[1] * (
    #             cas.tan(theta_rad) + c[2] * theta_rad
    #     )
    # )

    # Model 2
    c = (
        -0.04636,
        -0.3171
    )
    factor = np.exp(
        c[0] * (
                np.tand(theta * 0.999) + c[1] * np.radians(theta)
        )
    )

    # # Model 3
    # p1 = -21.74
    # p2 = 282.6
    # p3 = -1538
    # p4 = 1786
    # q1 = -923.2
    # q2 = 1456
    # x = theta_rad
    # factor = ((p1*x**3 + p2*x**2 + p3*x + p4) /
    #            (x**2 + q1*x + q2))

    return factor


def solar_flux(
        latitude: float,
        day_of_year: float,
        time: float,
        panel_azimuth_angle: float = 0,
        panel_tilt_angle: float = 0,
        scattering: bool = True
) -> float:
    """
    What is the solar flux on a horizontal surface for some given conditions?
    :param latitude: Latitude [degrees]
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :param time: Time since (local) solar noon [seconds]
    :param panel_azimuth_angle: The azimuth angle of the panel normal, in degrees. (0 degrees if pointing North and 90 if East)
    :param panel_tilt_angle: The angle between the panel normal and vertical, in degrees. (0 if horizontal and 90 if vertical)
    :param scattering: Boolean: include scattering effects at very low angles?

    :return: The solar flux on the panel, expressed in W/m^2.
    """
    return (
            solar_flux_outside_atmosphere_normal(day_of_year) *
            incidence_angle_function(
                latitude=latitude,
                day_of_year=day_of_year,
                time=time,
                panel_azimuth_angle=panel_azimuth_angle,
                panel_tilt_angle=panel_tilt_angle,
                scattering=scattering
            )
    )


def peak_sun_hours_per_day_on_horizontal(latitude, day_of_year, scattering=True):
    """
    How many hours of equivalent peak sun do you get per day?
    :param latitude: Latitude [degrees]
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :param time: Time since (local) solar noon [seconds]
    :param scattering: Boolean: include scattering effects at very low angles?
    :return:
    """
    times = np.linspace(0, 86400, 1000)
    dt = np.diff(times)
    normalized_fluxes = (
        # solar_flux_outside_atmosphere_normal(day_of_year) *
        incidence_angle_function(latitude, day_of_year, times, scattering)
    )
    sun_hours = np.sum(
        (normalized_fluxes[1:] + normalized_fluxes[:-1]) / 2 * dt
    ) / 3600

    return sun_hours


def length_day(latitude, day_of_year):
    """
    For what length of time is the sun above the horizon on a given day?

    :param latitude: Latitude [degrees]
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :return: Seconds of sunlight in a given day
    """
    dec = declination_angle(day_of_year)

    constant = -np.sind(dec) * np.sind(latitude) / (np.cosd(dec) * np.cosd(latitude))
    constant = np.clip(constant, -1, 1)

    sun_time_nondim = 2 * np.arccos(constant)
    sun_time = sun_time_nondim / (2 * np.pi) * 86400

    return sun_time


def mass_MPPT(
        power: float
) -> float:
    """
    Gives the estimated mass of a Maximum Power Point Tracking (MPPT) unit for solar energy
    collection. Based on regressions at AeroSandbox/studies/SolarMPPTMasses.

    Args:
        power: Power of MPPT [watts]

    Returns:
        Estimated MPPT mass [kg]
    """
    constant = 0.066343
    exponent = 0.515140
    return constant * power ** exponent


if __name__ == "__main__":
    pass
    # # circular flight path function test
    #
    # import matplotlib.pyplot as plt
    #
    # ## fun integration problem
    # airspeed = 30  # m/s
    # radius = 50000  # m
    # latitude = 60
    # day_of_year = 174
    # angle = 90
    # day = 60 * 60 * 24
    # # wind_speed = 25
    # # wind_direction = 180
    # # solar_flux = []
    # # ground_dist_list = [0]
    # # x_vector = []
    # # time_step = times[2] - times[1]
    # # groundspeed = 0
    # # for i, time in enumerate(times):
    # #     x = airspeed * time
    # #     x_vector.append(x)
    # #     vehicle_direction = x / (np.pi / 180) / radius + 90
    # #     heading_x = airspeed * np.sind(vehicle_direction) - wind_speed * np.sind(wind_direction)
    # #     heading_y = airspeed * np.cosd(vehicle_direction) - wind_speed * np.cosd(wind_direction)
    # #     groundspeed = np.sqrt(heading_x ** 2 + heading_y ** 2)
    # #     ground_dist = ground_dist_list[i-1] + groundspeed * time_step
    # #     ground_dist_list.append(ground_dist)
    # #     vehicle_heading = np.arctan2d(heading_y, heading_x)
    # #     panel_heading = vehicle_heading - 90
    # #     solar_flux.append(solar_flux_circular_flight_path(latitude, day_of_year, time, -angle, panel_heading) + solar_flux_circular_flight_path(latitude, day_of_year, time, angle, panel_heading) )
    #
    # # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # # plt.plot(times / 3600, solar_flux)
    # # plt.grid(True)
    # # plt.title("Solar Flux on Vertical as  Aircraft Completes a Circular Flight Path")
    # # plt.xlabel("Time after Solar Noon [hours]")
    # # plt.ylabel(r"Solar Flux [W/m$^2$]")
    # # plt.tight_layout()
    # # # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_horizontal")
    # # plt.show()
    # #
    # # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # # plt.plot(times / 3600, ground_dist_list[:-1])
    # # plt.grid(True)
    # # plt.title("Ground Distance Covered as Aircraft Completes a Circular Flight Path")
    # # plt.xlabel("Time after Solar Noon [hours]")
    # # plt.ylabel("Ground Distance [m]")
    # # plt.tight_layout()
    # # # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_horizontal")
    # # plt.show()
    #
    # # circular flight path test w optimization
    # import aerosandbox as asb
    #
    # opti = asb.Opti()
    # radius = 50000  # m
    # latitude = 60
    # day_of_year = 174
    # angle = 90
    # wind_speed = 20
    # wind_direction = 90
    #
    # n_timesteps_per_segment = 300
    # time_start = 0 * 3600
    # time_end = 24 * 3600
    #
    # time = np.linspace(
    #     time_start,
    #     time_end,
    #     n_timesteps_per_segment
    # )
    # time_periodic_start_index = 0
    # time_periodic_end_index = time.shape[0] - 1
    #
    # n_timesteps = time.shape[0]
    # hour = time / 3600
    #
    # # initialize variables
    # airspeed = opti.variable(
    #     n_vars=n_timesteps,
    #     init_guess=25,
    #     scale=10,
    # )
    # x = opti.variable(
    #     n_vars=n_timesteps,
    #     init_guess=0,
    #     scale=1e5,
    # )
    # heading_x = opti.variable(
    #     n_vars=n_timesteps,
    #     init_guess=90,
    #     scale=20,
    # )
    # heading_y = opti.variable(
    #     n_vars=n_timesteps,
    #     init_guess=180,
    #     scale=20,
    # )
    # vehicle_heading = opti.variable(
    #     n_vars=n_timesteps,
    #     init_guess=90,
    #     scale=20,
    # )
    # vehicle_direction = opti.variable(
    #     n_vars=n_timesteps,
    #     init_guess=80,
    #     scale=20,
    # )
    # groundspeed = opti.variable(
    #     n_vars=n_timesteps,
    #     init_guess=3,
    #     scale=1,
    # )
    #
    # # constraints
    # trapz = lambda x: (x[1:] + x[:-1]) / 2
    # dt = np.diff(time)
    # dx = np.diff(x)
    # opti.subject_to([
    #     airspeed >= 0,
    #     dx / 1e4 == trapz(groundspeed) * dt / 1e4,
    #     x[-1] >= 10000,
    #     vehicle_direction == x / (np.pi / 180) / radius + 90,
    #     # direction the vehicle must fly on to remain in the circular trajectory
    #     heading_x == airspeed * np.sind(vehicle_direction) - wind_speed * np.sind(wind_direction),
    #     # x component of heading vector
    #     heading_y == airspeed * np.cosd(vehicle_direction) - wind_speed * np.cosd(wind_direction),
    #     # y component of heading vector
    #     vehicle_heading == np.arctan2d(heading_y, heading_x),
    #     # actual directionality of the vehicle as modified by the wind speed and direction
    #     groundspeed ** 2 == heading_x ** 2 + heading_y ** 2,
    #     # speed of aircraft as measured from observer on the ground
    #     groundspeed >= 2,
    #     airspeed <= 40,
    #     airspeed >= 10,
    #     x > 0,
    #     x[0] == 0,
    # ])
    #
    # panel_heading = vehicle_heading - 90  # actual directionality of the solar panel
    #
    # solar_flux_on_vertical_left = solar_flux_circular_flight_path(
    #     latitude, day_of_year, time, 90, panel_heading, scattering=True,
    # )
    # solar_flux_on_vertical_right = solar_flux_circular_flight_path(
    #     latitude, day_of_year, time, -90, panel_heading, scattering=True,
    # )
    # solar_flux = solar_flux_on_vertical_left + solar_flux_on_vertical_right
    # solar_flux_total = np.sum(solar_flux_on_vertical_left + solar_flux_on_vertical_right)
    # opti.minimize(-solar_flux_total)
    # sol = opti.solve(
    #     max_iter=10000,
    # )
    #
    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # plt.plot(opti.value(time / 3600), opti.value(solar_flux))
    # plt.grid(True)
    # plt.title("Solar Flux on Vertical as  Aircraft Completes a Circular Flight Path")
    # plt.xlabel("Time after Solar Noon [hours]")
    # plt.ylabel(r"Solar Flux [W/m$^2$]")
    # plt.tight_layout()
    # # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_horizontal")
    # plt.show()
    #
    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # plt.plot(opti.value(time / 3600), opti.value(x))
    # plt.grid(True)
    # plt.title("Ground Distance Covered as Aircraft Completes a Circular Flight Path")
    # plt.xlabel("Time after Solar Noon [hours]")
    # plt.ylabel("Ground Distance [m]")
    # plt.tight_layout()
    # # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_horizontal")
    # plt.show()
    #
    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # plt.plot(opti.value(time / 3600), opti.value(groundspeed))
    # plt.grid(True)
    # plt.title("Groundspeed as Aircraft Completes a Circular Flight Path")
    # plt.xlabel("Time after Solar Noon [hours]")
    # plt.ylabel("Groundspeed [m/s]")
    # plt.tight_layout()
    # # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_horizontal")
    # plt.show()
    #
    # # Run some checks
    # import plotly.graph_objects as go
    #
    # latitudes = np.linspace(0, 80, 200)
    # day_of_years = np.arange(0, 365) + 1
    # times = np.linspace(0, 86400, 400)
    # #
    # headings = np.linspace(0, 180, 400)
    #
    # Times, Latitudes = np.meshgrid(times, latitudes, indexing="ij")
    # fluxes = np.array(solar_flux_on_horizontal(Latitudes, 244, Times))
    # fig = go.Figure(
    #     data=[
    #         go.Surface(
    #             x=Times / 3600,
    #             y=Latitudes,
    #             z=fluxes,
    #         )
    #     ],
    # )
    # fig.update_layout(
    #     scene=dict(
    #         xaxis=dict(title="Time after Solar Noon [hours]"),
    #         yaxis=dict(title="Latitude [deg]"),
    #         zaxis=dict(title="Solar Flux [W/m^2]"),
    #         camera=dict(
    #             eye=dict(x=-1, y=-1, z=1)
    #         )
    #     ),
    #     title="Solar Flux on Horizontal",
    # )
    # fig.show()
    ## test new function for off horizontal solar arrays
    # latitude = 60
    # angle = 90
    # Times, Headings = np.meshgrid(times, headings, indexing="ij")
    # fluxes = np.array(solar_flux_new(latitude, 174, times, headings, angle))
    # fig = go.Figure(
    #     data=[
    #         go.Surface(
    #             x=Times / 3600,
    #             y=Headings,
    #             z=fluxes,
    #         )
    #     ],
    # )
    # fig.update_layout(
    #     scene=dict(
    #         xaxis=dict(title="Time after Solar Noon [hours]"),
    #         yaxis=dict(title="Aircraft Heading [deg]"),
    #         zaxis=dict(title="Solar Flux [W/m^2]"),
    #         camera=dict(
    #             eye=dict(x=-1, y=-1, z=1)
    #         )
    #     ),
    #     title="Solar Flux on Vertical",
    # )
    # fig.show()
    # for heading in range(np.linspace(0, 180, 10)):
    #     fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    #     # lats_to_plot = [26, 49]
    #     lats_to_plot = np.linspace(0, 90, 7)
    #     colors = plt.cm.rainbow(np.linspace(0, 1, len(lats_to_plot)))[::-1]
    #     [
    #         plt.plot(
    #             times / 3600,
    #             solar_flux_on_new(lats_to_plot[i], 173, times),
    #             label="%iN Latitude" % lats_to_plot[i],
    #             color=colors[i],
    #             linewidth=3
    #         ) for i in range(len(lats_to_plot))
    #     ]
    #     plt.grid(True)
    #     plt.legend()
    #     plt.title("Solar Flux on a Vertical Surface (Summer Solstice)")
    #     plt.xlabel("Time after Solar Noon [hours]")
    #     plt.ylabel(r"Solar Flux [W/m$^2$]")
    #     plt.tight_layout()
    #     # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_vertical")
    #     plt.show()
    #
    # fig = go.Figure(
    #     data=[
    #         go.Contour(
    #             z=fluxes.T,
    #             x=times/3600,
    #             y=latitudes,
    #             colorbar=dict(
    #                 title="Solar Flux [W/m^2]"
    #             ),
    #             colorscale="Viridis",
    #         )
    #     ]
    # )
    # fig.update_layout(
    #     scene=dict(
    #         xaxis=dict(title="Hours after Solar Noon [hours]"),
    #         yaxis=dict(title="Latitude [deg]"),
    #     ),
    #     title="Solar Flux on Horizontal",
    #     xaxis_title="Time after Solar Noon [hours]",
    #     yaxis_title="Latitude [deg]",
    # )
    # fig.show()

    import matplotlib.pyplot as plt
    import seaborn as sns

    #
    # sns.set(font_scale=1)
    # latitudes = np.linspace(0, 80, 200)
    # day_of_years = np.arange(0, 365) + 1
    # times = np.linspace(0, 86400, 400)
    # latitude = 60
    # angle = 90
    # headings = np.linspace(0, 180, 400)
    #
    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # headings_to_plot = np.linspace(0, 180, 5)
    # colors = plt.cm.rainbow(np.linspace(0, 1, len(headings_to_plot)))[::-1]
    # [
    #     plt.plot(
    #         times / 3600,
    #         solar_flux_new(latitude,173, times, headings_to_plot[i], angle=0),
    #         label="%iN Headings" % headings_to_plot[i],
    #         color=colors[i],
    #         linewidth=3
    #     ) for i in range(len(headings_to_plot))
    # ]
    # plt.grid(True)
    # plt.legend()
    # plt.title("Solar Flux on a Horizontal Surface (Summer Solstice)")
    # plt.xlabel("Time after Solar Noon [hours]")
    # plt.ylabel(r"Solar Flux [W/m$^2$]")
    # plt.tight_layout()
    # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_horizontal")
    # plt.show()
    #
    #
    sns.set(font_scale=1)

    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # lats_to_plot = [26, 49]
    # lats_to_plot = np.linspace(0, 90, 7)
    # colors = plt.cm.rainbow(np.linspace(0, 1, len(lats_to_plot)))[::-1]
    # [
    #     plt.plot(
    #         times / 3600,
    #         solar_flux_on_horizontal(lats_to_plot[i], 173, times),
    #         label="%iN Latitude" % lats_to_plot[i],
    #         color=colors[i],
    #         linewidth=3
    #     ) for i in range(len(lats_to_plot))
    # ]
    # plt.grid(True)
    # plt.legend()
    # plt.title("Solar Flux on a Horizontal Surface (Summer Solstice)")
    # plt.xlabel("Time after Solar Noon [hours]")
    # plt.ylabel(r"Solar Flux [W/m$^2$]")
    # plt.tight_layout()
    # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_horizontal")
    # plt.show()
    # # #
    # sns.set(font_scale=1)
    #
    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # lats_to_plot = [26, 49]
    # lats_to_plot = np.linspace(0, 90, 7)
    # colors = plt.cm.rainbow(np.linspace(0, 1, len(lats_to_plot)))[::-1]
    # [
    #     plt.plot(
    #         times / 3600,
    #         solar_flux_on_vertical(lats_to_plot[i], 173, times),
    #         label="%iN Latitude" % lats_to_plot[i],
    #         color=colors[i],
    #         linewidth=3
    #     ) for i in range(len(lats_to_plot))
    # ]
    # plt.grid(True)
    # plt.legend()
    # plt.title("Solar Flux on a Vertical Surface (Summer Solstice)")
    # plt.xlabel("Time after Solar Noon [hours]")
    # plt.ylabel(r"Solar Flux [W/m$^2$]")
    # plt.tight_layout()
    # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_vertical")
    # plt.show()
    #
    # sns.set(font_scale=1)
    #
    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # lats_to_plot = [26, 49]
    # lats_to_plot = np.linspace(0, 90, 7)
    # colors = plt.cm.rainbow(np.linspace(0, 1, len(lats_to_plot)))[::-1]
    # [
    #     plt.plot(
    #         times / 3600,
    #         solar_flux_on_angle(10, lats_to_plot[i], 173, times),
    #         label="%iN Latitude" % lats_to_plot[i],
    #         color=colors[i],
    #         linewidth=3
    #     ) for i in range(len(lats_to_plot))
    # ]
    # plt.grid(True)
    # plt.legend()
    # plt.title("Solar Flux on a Angle 10 Degrees off Vertical Surface (Summer Solstice)")
    # plt.xlabel("Time after Solar Noon [hours]")
    # plt.ylabel(r"Solar Flux [W/m$^2$]")
    # plt.tight_layout()
    # plt.savefig("/Users/annickdewald/Desktop/Thesis/Photos/solar_angle")
    # plt.show()
    #
    # # Check scattering factor
    # elevations = np.linspace(-10,90,800)
    # scatter_factors = scattering_factor(elevations)
    #
    # import matplotlib.pyplot as plt
    # import matplotlib.style as style
    # import seaborn as sns
    # sns.set(font_scale=1)
    # fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    # plt.plot(elevations, scatter_factors,'.-')
    # plt.xlabel(r"Elevation Angle [deg]")
    # plt.ylabel(r"Scattering Factor")
    # plt.title(r"Scattering Factor")
    # plt.tight_layout()
    # # plt.legend()
    # # plt.savefig("C:/Users/User/Downloads/temp.svg")
    # plt.show()



# Lengths (equivalents in meters)
meter = 1
inch = 0.0254
foot = 12 * inch
yard = 3 * foot
mile = 5280 * foot
naut_mile = 1852

# Masses (equivalents in kilograms)
kg = 1
gram = 0.001
slug = 14.59390
lbm = 0.45359237
short_ton = 2000 * lbm
long_ton = 2240 * lbm
oz = lbm / 16

# Time (equivalents in seconds)
second = 1
minute = 60
hour = 60 * minute
day = 86400
week = 7 * day
year = 365.25 * day
month = year / 12

# Force (equivalents in Newtons)
newton = 1
lbf = slug * foot

# Speed (equivalents in m/s)
knot = 1.852 * 1000 / hour
mph = mile / hour
fps = foot / 1

# Volume (equivalents in m^3)
liter = 0.001

# SI prefixes
yocto = 1e-24
zepto = 1e-21
atto = 1e-18
femto = 1e-15
pico = 1e-12
nano = 1e-9
micro = 1e-6
milli = 1e-3
centi = 1e-2
deci = 1e-1
deka = 1e1
hecto = 1e2
kilo = 1e3
mega = 1e6
giga = 1e9
tera = 1e12
peta = 1e15
exa = 1e18
zetta = 1e21
yotta = 1e24



import aerosandbox as asb
import aerosandbox.numpy as np
from typing import List
import copy
import pytest

vector = [1, 2, 3]
dyn = asb.DynamicsRigidBody3DBodyEuler(
    x_e=0,
    y_e=0,
    z_e=0,
    u_b=1,
    v_b=0,
    w_b=0,
    phi=0,
    theta=0,
    psi=0,
    p=0,
    q=0,
    r=0,
)


def chain_conversion(
        axes: List[str] = ["geometry", "body", "geometry"]
):
    x, y, z = copy.deepcopy(vector)
    for from_axes, to_axes in zip(axes, axes[1:]):
        x, y, z = dyn.convert_axes(
            x_from=x,
            y_from=y,
            z_from=z,
            from_axes=from_axes,
            to_axes=to_axes
        )
    return np.array(vector) == pytest.approx(np.array([x, y, z]))


def test_basic():
    assert chain_conversion()


def test_geometry():
    assert chain_conversion(["body", "geometry", "body"])


def test_stability():
    assert chain_conversion(["body", "stability", "body"])


def test_wind():
    assert chain_conversion(["body", "wind", "body"])


def test_earth():
    assert chain_conversion(["body", "earth", "body"])


def test_cycle():
    assert chain_conversion([
        "body",
        "geometry",
        "stability",
        "wind",
        "body",
        "wind",
        "stability",
        "geometry",
        "body",
        "geometry",
        "wind",
        "geometry",
        "stability",
        "body",
    ])


if __name__ == '__main__':
    pytest.main()
    chain_conversion()



import matplotlib.pyplot as plt


def figure3d(*args, orthographic=True, **kwargs):
    """
    Creates a new 3D figure. Args and kwargs are passed into matplotlib.pyplot.figure().

    Returns: (fig, ax)

    """
    fig = plt.figure(*args, **kwargs)

    axes_args = dict(
        projection='3d'
    )
    if orthographic:
        axes_args["proj_type"] = 'ortho'

    ax = plt.axes(**axes_args)
    return fig, ax


if __name__ == '__main__':
    figure3d()



from aerosandbox.tools.inspect_tools import *
import pytest
import inspect


def test_function_argument_names_from_source_code():
    tests = {  # Pairs of {input: expected_output}
        "f(a, b)"              : ['a', 'b'],
        "f(a,b)"               : ['a', 'b'],
        "f(\na,\nb)"           : ['a', 'b'],
        "g = f(a, b)"          : ['a', 'b'],
        "g.h = f(a, b)"        : ['a', 'b'],
        "g.h() = f(a, b)"      : ['a', 'b'],
        "g.h(i=j) = f(a, b)"   : ['a', 'b'],
        "f(a, b) + g(h)"       : ['a', 'b'],
        "f(a: int, b: float())": ['a', 'b'],
        "f(a, b).g(c, d)"      : ['a', 'b'],
        "f(a(b), c)"           : ['a(b)', 'c'],
        "f(a(b,c), d)"         : ['a(b,c)', 'd'],
        "f({a:b}, c)"          : ['{a:b}', 'c'],
        "f(a[b], c)"           : ['a[b]', 'c'],
        "f({a:b, c:d}, e)"     : ['{a:b,c:d}', 'e'],
        "f({a:b,\nc:d}, e)"    : ['{a:b,c:d}', 'e'],
        "f(dict(a=b,c=d), e)"  : ['dict(a=b,c=d)', 'e'],
        "f(a=1, b=2)"          : ['a=1', 'b=2'],
        "3 + 5"                : ValueError,
        ""                     : ValueError,
    }

    for input, expected_output in tests.items():

        ### If you're expecting an error, make sure it gets raised
        if inspect.isclass(expected_output) and issubclass(expected_output, Exception):
            with pytest.raises(expected_output):
                get_function_argument_names_from_source_code(input)

        ### If you're expecting a specific output, make sure you get that
        else:
            assert get_function_argument_names_from_source_code(input) == expected_output


if __name__ == '__main__':
    test_function_argument_names_from_source_code()
    pytest.main()



"""
A series of utilities for working with CSV data extracted from WebPlotDigitizer.

https://automeris.io/WebPlotDigitizer/
https://github.com/ankitrohatgi/WebPlotDigitizer

"""
import numpy as np
from typing import Dict


def read_webplotdigitizer_csv(filename) -> Dict[str, np.ndarray]:
    delimiter = ","
    with open(filename, "r") as f:
        lines = f.readlines()

    titles = lines[0].split(delimiter)[::2]

    def string_to_float(s: str) -> float:
        try:
            return float(s)
        except ValueError:
            return np.NaN

    data = [
        [string_to_float(item) for item in line.split(delimiter)]
        for line in lines[2:]
    ]

    data = np.array(data, dtype=float)

    def remove_nan_rows(a: np.ndarray) -> np.ndarray:
        nan_rows = np.any(np.isnan(a), axis=1)
        return a[~nan_rows, :]

    return {
        title: remove_nan_rows(data[:, 2 * i:2 * i + 2])
        for i, title in enumerate(titles)
    }



from aerosandbox import AeroSandboxObject
from aerosandbox.geometry.common import *
from typing import List, Dict, Any, Union
import aerosandbox.geometry.mesh_utilities as mesh_utils


class Airplane(AeroSandboxObject):
    """
    Definition for an airplane.

    Anatomy of an Airplane:

        An Airplane consists chiefly of a collection of wings and fuselages. These can be accessed with
        `Airplane.wings` and `Airplane.fuselages`, which gives a list of those respective components. Each wing is a
        Wing object, and each fuselage is a Fuselage object.
    """

    def __init__(self,
                 name: str = "Untitled",
                 xyz_ref: Union[np.ndarray, List] = np.array([0, 0, 0]),
                 wings: List['Wing'] = None,
                 fuselages: List['Fuselage'] = None,
                 s_ref: float = None,
                 c_ref: float = None,
                 b_ref: float = None,
                 analysis_specific_options: Dict[type, Dict[str, Any]] = None
                 ):
        """
        Defines a new airplane.

        Args:

            name: Name of the airplane [optional]. It can help when debugging to give the airplane a sensible name.

            xyz_ref: An array-like that gives the x-, y-, and z- reference point of the airplane, used when computing
            moments and stability derivatives. Generally, this should be the center of gravity.

                # In a future version, this will be deprecated and replaced with asb.MassProperties.

            wings: A list of Wing objects that are a part of the airplane.

            fuselages: A list of Fuselage objects that are a part of the airplane.

            s_ref: Reference area. If undefined, it's set from the area of the first Wing object. # Note: will be deprecated

            c_ref: Reference chord. If undefined, it's set from the mean aerodynamic chord of the first Wing object. # Note: will be deprecated

            b_ref: Reference span. If undefined, it's set from the span of the first Wing object. # Note: will be deprecated

            analysis_specific_options: Analysis-specific options are additional constants or modeling assumptions
            that should be passed on to specific analyses and associated with this specific geometry object.

                This should be a dictionary where:

                    * Keys are specific analysis types (typically a subclass of asb.ExplicitAnalysis or
                    asb.ImplicitAnalysis), but if you decide to write your own analysis and want to make this key
                    something else (like a string), that's totally fine - it's just a unique identifier for the
                    specific analysis you're running.

                    * Values are a dictionary of key:value pairs, where:

                        * Keys are strings.

                        * Values are some value you want to assign.

                This is more easily demonstrated / understood with an example:

                >>> analysis_specific_options = {
                >>>     asb.AeroBuildup: dict(
                >>>         include_wave_drag=True,
                >>>     )
                >>> }

        """
        ### Set defaults
        if wings is None:
            wings: List['Wing'] = []
        if fuselages is None:
            fuselages: List['Fuselage'] = []
        if analysis_specific_options is None:
            analysis_specific_options = {}

        ### Initialize
        self.name = name
        self.xyz_ref = np.array(xyz_ref)
        self.wings = wings
        self.fuselages = fuselages
        self.analysis_specific_options = analysis_specific_options

        ### Assign reference values
        try:
            main_wing = self.wings[0]
            if s_ref is None:
                s_ref = main_wing.area()
            if c_ref is None:
                c_ref = main_wing.mean_aerodynamic_chord()
            if b_ref is None:
                b_ref = main_wing.span()
        except IndexError:
            s_ref = 1
            c_ref = 1
            b_ref = 1
        self.s_ref = s_ref
        self.c_ref = c_ref
        self.b_ref = b_ref

    def __repr__(self):
        n_wings = len(self.wings)
        n_fuselages = len(self.fuselages)
        return f"Airplane '{self.name}' " \
               f"({n_wings} {'wing' if n_wings == 1 else 'wings'}, " \
               f"{n_fuselages} {'fuselage' if n_fuselages == 1 else 'fuselages'})"

    # TODO def add_wing(wing: 'Wing') -> None

    def mesh_body(self,
                  method="quad",
                  thin_wings=False,
                  stack_meshes=True,
                  ):
        """
        Returns a surface mesh of the Airplane, in (points, faces) format. For reference on this format,
        see the documentation in `aerosandbox.geometry.mesh_utilities`.

        Args:

            method:

            thin_wings: Controls whether wings should be meshed as thin surfaces, rather than full 3D bodies.

            stack_meshes: Controls whether the meshes should be merged into a single mesh or not.

                * If True, returns a (points, faces) tuple in standard mesh format.

                * If False, returns a list of (points, faces) tuples in standard mesh format.

        Returns:

        """
        if thin_wings:
            wing_meshes = [
                wing.mesh_thin_surface(
                    method=method,
                )
                for wing in self.wings
            ]
        else:
            wing_meshes = [
                wing.mesh_body(
                    method=method,
                )
                for wing in self.wings
            ]

        fuse_meshes = [
            fuse.mesh_body(
                method=method
            )
            for fuse in self.fuselages
        ]

        meshes = wing_meshes + fuse_meshes

        if stack_meshes:
            points, faces = mesh_utils.stack_meshes(*meshes)
            return points, faces
        else:
            return meshes

    def draw(self,
             backend: str = "pyvista",
             thin_wings: bool = False,
             show: bool = True,
             show_kwargs=None,
             ):
        """

        Args:

            backend: One of:
                * "plotly" for a Plot.ly backend
                * "pyvista" for a PyVista backend
                * "trimesh" for a trimesh backend

            thin_wings: A boolean that determines whether to draw the full airplane (i.e. thickened, 3D bodies), or to use a
            thin-surface representation for any Wing objects.

            show: Should we show the visualization, or just return it?

        Returns: The plotted object, in its associated backend format. Also displays the object if `show` is True.

        """
        if show_kwargs is None:
            show_kwargs = {}

        if backend == "plotly":

            points, faces = self.mesh_body(method="quad", thin_wings=thin_wings)

            from aerosandbox.visualization.plotly_Figure3D import Figure3D
            fig = Figure3D()
            for f in faces:
                fig.add_quad((
                    points[f[0]],
                    points[f[1]],
                    points[f[2]],
                    points[f[3]],
                ), outline=True)
                show_kwargs = {
                    "show": show,
                    **show_kwargs
                }
            return fig.draw(**show_kwargs)
        elif backend == "pyvista":

            points, faces = self.mesh_body(method="quad")

            import pyvista as pv
            fig = pv.PolyData(
                *mesh_utils.convert_mesh_to_polydata_format(points, faces)
            )
            show_kwargs = {
                "show_edges": True,
                "show_grid" : True,
                **show_kwargs,
            }
            if show:
                fig.plot(**show_kwargs)
            return fig
        elif backend == "trimesh":

            points, faces = self.mesh_body(method="tri")

            import trimesh as tri
            fig = tri.Trimesh(points, faces)
            if show:
                fig.show(**show_kwargs)
            return fig
        else:
            raise ValueError("Bad value of `backend`!")

    def is_entirely_symmetric(self):
        """
        Returns a boolean describing whether the airplane is geometrically entirely symmetric across the XZ-plane.
        :return: [boolean]
        """
        for wing in self.wings:
            if not wing.is_entirely_symmetric():
                return False

        # TODO add in logic for fuselages

        return True

    def aerodynamic_center(self, chord_fraction: float = 0.25):
        """
        Computes the location of the aerodynamic center of the wing.
        Uses the generalized methodology described here:
            https://core.ac.uk/download/pdf/79175663.pdf

        Args:
            chord_fraction: The position of the aerodynamic center along the MAC, as a fraction of MAC length.
                Typically, this value (denoted `h_0` in the literature) is 0.25 for a subsonic wing.
                However, wing-fuselage interactions can cause a forward shift to a value more like 0.1 or less.
                Citing Cook, Michael V., "Flight Dynamics Principles", 3rd Ed., Sect. 3.5.3 "Controls-fixed static stability".
                PDF: https://www.sciencedirect.com/science/article/pii/B9780080982427000031

        Returns: The (x, y, z) coordinates of the aerodynamic center of the airplane.
        """
        wing_areas = [wing.area(type="projected") for wing in self.wings]
        ACs = [wing.aerodynamic_center() for wing in self.wings]

        wing_AC_area_products = [
            AC * area
            for AC, area in zip(
                ACs,
                wing_areas
            )
        ]

        aerodynamic_center = sum(wing_AC_area_products) / sum(wing_areas)

        return aerodynamic_center
